{
  Associative arrays for Pascal
  by Bidok (szkuwa [at] o2 [dot] pl)

  version 1.4

  This code is free only for non-commercial usage :)
  
  History:
  ] 1.4 - No variants in TCustomAssocArray class..
          this should speed up code a little
        - now, arrays can be accessed only by string key.. if you want others
          like tab[123] or tab[12.34] you have to write own class
          (maybe i'll add some later)
  ] 1.3 - minor changes in code for delphi support (maybe more is needed..)
        - added new properties: ChildCount, Length and Rank
        - function HaveChilds
  ] 1.2 - rewrite almost all code
        - support for unlimited diemsions arrays
        - now you can choose which type of data you want to store in array
          (string, integer, pointer, variant)
        - or you can write own class based on TCustomAssocArray and use as many
          field as you want ;)
  ] 1.1 - added variants support
        - you can choose (pointers/variants) with $DEFINE switch
  ] 1.0 - first release, work only with pointers
        - arrays have only one diemension
}
unit assocabase;
{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}

interface

type
  TCustomAssocArrayClass = class of TCustomAssocArray;
  TCustomAssocArray = class
  private
    AList: TCustomAssocArray;
    ALast: TCustomAssocArray;
    FName: string;
    FNext: TCustomAssocArray;
    FParent: TCustomAssocArray;
    FLevel: Integer;
    FMagicClass: TCustomAssocArrayClass;
    
    procedure RemoveItemFromList(AName: string);
  public 
    constructor Create; overload;
    constructor Create(AOwner: TCustomAssocArray); overload;
    destructor Destroy; override;
    procedure Clear;

    function GetItem(AName: string): TCustomAssocArray; virtual;
    function ForEach(var Item: TCustomAssocArray): Boolean;
    function ForEachLevel(var Item: TCustomAssocArray): Boolean;
    function GetLength: Integer;
    function GetRank: Integer;
    function HaveChilds: Boolean;
    function GetChildCount: Integer;

    property Items[const AName: string]: TCustomAssocArray read GetItem; default;
    property Name: string read FName;
    property Next: TCustomAssocArray read FNext;
    property Level: Integer read FLevel;
    property MagicClass: TCustomAssocArrayClass read FMagicClass write FMagicClass;
    property Parent: TCustomAssocArray read FParent;
    property Length: Integer read GetLength;
    property Rank: Integer read GetRank;
    property ChildCount: Integer read GetChildCount;
  end;

implementation

{ TCustomAssocArray }
constructor TCustomAssocArray.Create;
begin
  inherited Create;
  AList := nil;
  ALast := nil;
  FParent := nil;
  FLevel := 0;
  FName := '';
  FMagicClass := nil; // AV: 
     // TCustomAssocArray nie moze byc wykorzystywane bezposrednio...
end;

constructor TCustomAssocArray.Create(AOwner: TCustomAssocArray);
begin
  inherited Create;
  AList := nil;
  ALast := nil;
  FParent := AOwner;
  FName := '';
  FLevel := AOwner.Level + 1;
  if AOwner.FMagicClass <> nil then
    FMagicClass := AOwner.FMagicClass else
    FMagicClass := nil; // AV ;]
end;

destructor TCustomAssocArray.Destroy;
begin
  Clear;
  
  if FParent <> nil then
     FParent.RemoveItemFromList(Name);

  inherited Destroy;
end;

procedure TCustomAssocArray.Clear;
var
  Temp1, Temp2: TCustomAssocArray;
begin
  if (not Assigned(AList)) or (AList = nil) then Exit;
  
  Temp1 := AList;
  while Assigned(Temp1) do
  begin
    Temp2 := Temp1.Next;
    Temp1.Free;
    Temp1 := Temp2;
  end;
  AList := nil;
  ALast := nil;
end;

function TCustomAssocArray.GetItem(AName: string): TCustomAssocArray;
var
  Temp: TCustomAssocArray;
begin
  Result := nil;
  
  Temp := AList;
  while (Temp <> nil) and (Temp.Name <> AName) do
    Temp := Temp.Next;
    
  if Temp = nil then
  begin
    Temp := FMagicClass.Create(Self);
    Temp.FName := AName;
    if AList = nil then
    begin
      AList := Temp;
      ALast := Temp;
    end else
    begin
      ALast.FNext := Temp;
      ALast := Temp;
    end;
  end;
  Result := Temp;
end;

procedure TCustomAssocArray.RemoveItemFromList(AName: string);
var
  Temp1, Temp2: TCustomAssocArray;
begin
  if AList = nil then Exit;
  
  Temp1 := AList;
  Temp2 := nil;
  while (Temp1 <> nil) and (Temp1.Name <> AName) do
  begin
    Temp2 := Temp1;
    Temp1 := Temp1.Next;
  end;
  
  if Temp1 <> nil then
  begin
    if Temp2 <> nil then
       Temp2.FNext := Temp1.FNext else
       AList := Temp1.Next;
  end;
end;

function TCustomAssocArray.GetLength: Integer;
var
  Item: TCustomAssocArray;
begin
  Item := nil;
  Result := 1;
  while ForEachLevel(Item) do
    Result := Result + 1;
end;

function TCustomAssocArray.GetRank: Integer;
var
  Item: TCustomAssocArray;
begin
  Item := nil;
  Result := 0;
  while ForEachLevel(Item) do
    if Item.Level > Result then Result := Item.Level-Level;
end;

function TCustomAssocArray.HaveChilds: Boolean;
begin
  Result := (AList <> nil) and Assigned(AList);
end;

function TCustomAssocArray.GetChildCount: Integer;
var
  Item: TCustomAssocArray;
begin
  Item := nil;
  Result := 0;
  while ForEach(Item) do
    Result := Result + 1;
end;

function TCustomAssocArray.ForEach(var Item: TCustomAssocArray): Boolean;
begin
  Result := False;
  if not Assigned(AList) or (AList = nil) then Exit;
  
  if (Item = nil) or (not Assigned(Item)) then
  begin
    Item := AList;
    Result := True;
    Exit;
  end;
  
  if (Item.Next <> nil) and (Assigned(Item.Next)) then
  begin
    Item := Item.Next;
    Result := True;
  end;
end;

function TCustomAssocArray.ForEachLevel(var Item: TCustomAssocArray): Boolean;
  function BackInTree(It: TCustomAssocArray): TCustomAssocArray;
  begin
    Result := It;
    while (Result <> nil) and (Result <> Self) and (Result.Next = nil) do
      Result := Result.Parent;
      
    if (Result = Self) then Result := nil;
    if (Result <> nil) then Result := Result.Next;
  end;
begin
  Result := False;
  if ((Item = nil) or (not Assigned(Item))) 
     and ((AList <> nil) and Assigned(AList)) then
  begin
    Item := AList;
    Result := True;
    Exit;
  end;
  
  if (Item <> nil) and Assigned(Item) and
     (Item.AList <> nil) and Assigned(Item.AList) then
  begin
    Item := Item.AList;
    Result := True;
    Exit;
  end;
  
  if (Item <> nil) and Assigned(Item) and
     (Item.AList = nil) and (Item.Next <> nil) and Assigned(Item.Next) then
  begin
    Item := Item.Next;
    Result := True;
    Exit;
  end;
  
  if (Item <> nil) and Assigned(Item) and
     (Item.Parent <> nil) and Assigned(Item.Parent) and 
     (Item.Next = nil) then
  begin
    Item := BackInTree(Item);
    Result := (Item <> nil);
    Exit;
  end;
end;

end.
